import type { ConfigEnv, ProxyOptions, UserConfig } from 'vite'
import { loadEnv } from 'vite'
import { resolve } from 'node:path'
import dayjs from 'dayjs'
import { isObject, isString } from '@gx-design-vue/pro-utils'

import { configManualChunk } from './build/optimizer'
import { createVitePlugins } from './build/plugin'
import { generateModifyVars } from './build/generateModifyVars'

import pkg from './package.json'

type ProxyTargetList = ProxyOptions & { rewrite: (path: string) => string }

const proxyTarget: string | Record<string, any> = 'http://127.0.0.1:3000'

const { dependencies, devDependencies, name, version } = pkg

const __APP_INFO__ = {
  pkg: { dependencies, devDependencies, name, version },
  lastBuildTime: dayjs().format('YYYY-MM-DD HH:mm:ss')
}

function pathResolve(dir: string) {
  return resolve(process.cwd(), '.', dir)
}

function createProxy(prefix) {
  const ret = {
    dev: {},
    test: {},
    pre: {}
  }

  if (isObject(proxyTarget)) {
    Object.keys(proxyTarget).forEach((prefix) => {
      const proxy = {
        target: `${proxyTarget[prefix]}`,
        changeOrigin: true,
        ws: true,
        rewrite: path => path.replace(new RegExp(`^${prefix}`), '')
      } as ProxyTargetList

      ret.dev[prefix] = proxy
      ret.test[prefix] = proxy
      ret.pre[prefix] = proxy
    })
  } else if (isString(proxyTarget)) {
    const proxy = {
      target: `${proxyTarget}`,
      changeOrigin: true,
      ws: true,
      rewrite: path => path.replace(new RegExp(`^${prefix}`), '')
    }
    ret.dev[prefix] = proxy
    ret.test[prefix] = proxy
    ret.pre[prefix] = proxy
  }

  return ret
}

function wrapperEnv(envConf: RecordType): ViteEnv {
  const ret: any = {}

  for (const envName of Object.keys(envConf)) {
    let realName = envConf[envName].replace(/\\n/g, '\n')
    realName = realName === 'true' ? true : realName === 'false' ? false : realName

    ret[envName] = realName
    process.env[envName] = realName
  }
  return ret
}

export default ({ mode, command }: ConfigEnv): UserConfig => {
  const root = process.cwd()

  const env = loadEnv(mode, root)

  // The boolean type read by loadEnv is a string. This function can be converted to boolean type
  const viteEnv = wrapperEnv(env)

  const { VITE_BASE_URL, VITE_DROP_CONSOLE } = viteEnv

  const isBuild = command === 'build'

  return {
    base: './',
    resolve: {
      alias: [
        {
          find: '@',
          replacement: pathResolve('src') + '/'
        },
        {
          find: '@gx-mock',
          replacement: pathResolve('mock') + '/'
        },
        {
          find: '@gx-vuex',
          replacement: pathResolve('src/store') + '/index.ts'
        },
        {
          find: '@gx-mobile/hooks',
          replacement: pathResolve('src/hooks')
        },
        { find: /^~/, replacement: '' }
      ]
    },
    server: {
      host: true,
      port: 3000,
      proxy: createProxy(VITE_BASE_URL)
    },
    plugins: createVitePlugins(viteEnv, isBuild),
    define: {
      __INTLIFY_PROD_DEVTOOLS__: false,
      __APP_INFO__: JSON.stringify(__APP_INFO__)
    },
    esbuild: {
      drop: VITE_DROP_CONSOLE ? [ 'console', 'debugger' ] : []
    },
    build: {
      assetsDir: 'assets',
      chunkSizeWarningLimit: 2500,
      rollupOptions: {
        output: {
          manualChunks: configManualChunk as any
        }
      }
    },
    css: {
      modules: {
        generateScopedName: 'gx-mobile-[local]-[hash:base64:5]'
      },
      preprocessorOptions: {
        less: {
          modifyVars: generateModifyVars(),
          javascriptEnabled: true
        }
      }
    }
  }
}
